{% extends "base.html" %}

{% block style %}
{% endblock style %}

{% block body %}

<script>
    // Remove the stored cookie named 'ceDashStoredHost'
    document.addEventListener("DOMContentLoaded", function() {
        document.cookie = "ceDashStoredHost=; expires=Thu, 01 Jan 1970 00:00:00 UTC; path=/;";
    });
</script>

<div class="container-fluid">

<div class="row">
<div class="col">
<p style="text-align: center">
All OSG-hosted HCondor CEs appear below.  Clicking on the <i>name</i> will open a browser tab with the Dashboard for that CE. Clicking on a table column name will sort that column.  
</p>
</div>
</div>
        <div class="row">
            <div class="col" id="ce-table">
                <script>
                    new HTCondorView({
                        dst_id: "ce-table",
						data_url: "/data/ce_landing",
                        {% set graph_query %}
                            group=facility_name,name,health,healthInfo,hosted,active;
                            &filter=hosted=1
                            &filter=active=1
                            &unselect=hosted,active
                            &order=facility_name,name
                            &intensify=health
                            &rename=facility_name=institution
                        {% endset %}
                        graph_query: "{{ graph_query | replace('\n', '') | replace(' ', '') }}",
                        has_edit_menu: false,
                    });
                </script>
            </div>
        </div>


</div>

{% endblock body %}
