/*https://stackoverflow.com/questions/4086107/fixed-page-header-overlaps-in-page-anchors*/
/*
*[id]:before {
    display: block;
    content: " ";
    margin-top: -75px;
    height: 75px;
    visibility: hidden;
}
*/
html {
    position: relative;
    min-height: 100%;
}

body {
    font-family: "Helvetica Neue", Helvetica, sans-serif;
    line-height: 1.5em;
    margin-bottom: 100px;
}

nav {
    margin-bottom: 10px;
}

.navbar {
    justify-content: space-between;
}

.navbar-collapse {
    flex-grow: 0;
}

.navbar-brand.abs
{
	/* position: absolute; */
	width: auto;
	left: 50%;
    top: 0.5em;
	/* transform: translateX(-50%); */
	text-align: center;
	align-items: center;
	font-size: 1.5rem;
	font-weight: bold;
	color: white;
    margin-left: 16px;
    margin-right: 16px;
}

.navbar-nav.site-link {
    font-size: 1rem; 
    font-weight: 400;
    padding: 0 8px;
    display: flex;
    flex-direction: column;
    justify-content: center;


    .nav-link {
        padding: 0;
        color: lightgray;

        i {
            font-size: 2rem;
            align-self: center;
        }

        &:hover {
            color: white;
        }
    }
}

h1 {
    font-family: "HelveticaNeue-Light", "HelveticaNeueLight", "Helvetica Neue Light", "Helvetica Neue", sans-serif;
    font-weight: 300;
    text-align: center;
    font-size: 250%;
}

.htcv-sidenav {
    height: 80vh;
    overflow-y: auto;
    margin-top: 20px;
}

.htcv-sidenav ul {
    list-style-type: none;
}

.htcv-sidenav ul.lvl1 {
    padding-left: 0em;
}

.htcv-sidenav ul.lvl2 {
    padding-left: 1.5em;
}

.htcondorview div.editmenu {
    position: absolute;
    right: 0;
    z-index: 20;
    background: #eee;
    visibility: hidden;
    line-height: 0.9em;
}

body .htcondorview div:hover > div.editmenu {
    visibility: visible;
}

.htcondorview div.editmenu a {
    text-decoration: none;
    padding: 2px;
}

.htcondorview div.editmenu a:hover {
    background: #aaa;
}

.htcondorview .vizchart {
    width: 100%;
    height: 100%;
    text-align: center;
    height: auto;
}

.htcondorview .graph {
    /* Allow the edit menu to be absolutely positioned inside */
    position: relative;
}

.htcondorview #vizlog {
    white-space: pre;
}

.htcondorview .vizstatus {
    position: absolute;
    z-index: 4;
    width: 100%;
    text-align: center;
    padding-top: 1em;
}

.htcondorview #statussub {
    padding-left: 1em;
    color: #aaa;
}

.htcondorview .vizstep {
    margin-top: 2em;
/* display: none; */
}

.htcondorview .vizstep .text {
    font-family: mono;
    font-weight: bold;
}

.htcondorview .vizstep .grid {
    margin-left: 3em;
    max-height: 10em;
    width: 50%;
    overflow: auto;
}

.htcondorview .dy-chart {
    border-bottom: 3px solid black;
}

.htcondorview .dy-chart-odd {
    background-color: #eee;
}

.htcondorview .dy-line {
    background-color: rgba(0, 0, 0, 0.2);
    position: absolute;
    pointer-events: none; /* don't capture double clicks, etc. */
}

.htcondorview .dy-xline {
    height: 100%;
    width: 3px;
    top: 0;
}

.htcondorview .dy-yline {
    height: 3px;
    width: 100%;
    left: 0;
}

.htcondorview .editlink {
    float: right;
    font-size: 80%;
}

.htcondorview button {
    box-shadow: 1px 1px 0px 0px #CF866C inset;
    background: #D0451B linear-gradient(to bottom, #D0451B 5%, #BC3315 100%) repeat scroll 0% 0%;
    border-radius: 6px;
    border: 1px solid #942911;
    display: inline-block;
    cursor: pointer;
    color: #FFF;
    padding: 0.3em 1em;
    text-decoration: none;
    text-shadow: 0px 1px 0px #854629;
}

.htcondorview .download-link {
    font-size: 80%;
    text-align: center;
}

.htcondorviewranged input[type="date"].range_input {
    width: 10em;
}

.htcondorviewranged input[type="time"].range_input {
    width: 8em;
}

.htcondorviewranged div.tabs {
    text-align: center;
}

.htcondorviewranged div.range {
    text-align: center;
}

div .tabs {
    position: relative;
    z-index: 100;
}

div .range {
    position: relative;
    z-index: 100;
}

ul.radio-tabs {
    display: inline;
    margin: 0;
    padding: 0;
    z-index: 1000;
}

ul.radio-tabs li {
    display: inline-block;
}

ul.radio-tabs li input[type="radio"] {
    display: none;
}

ul.radio-tabs li label {
    display: block;
    border: 1px solid black;
    background: #ededed;
    padding: .1em .8em;
    cursor: pointer;
}

ul.radio-tabs li:first-child label {
    border-top-left-radius: .6em;
    border-bottom-left-radius: 0.6em;
}

ul.radio-tabs li:last-child label {
    border-top-right-radius: 0.6em;
    border-bottom-right-radius: 0.6em;
}

ul.radio-tabs li input[type="radio"]:checked + label {
    background: white;
    cursor: inherit;
}

/* For tables, left align text */
.google-visualization-table-td {
    text-align: left;
}

/*  Prevent the browser page from scrolling when 
    selecting between Day/Week/Month by giving
    a big pre-defined height to the table under
    the graph */
.detail_table {
    height: 99vh;
}

/* For HTML tooltips, get rid of all the padding */
.google-visualization-tooltip {
    padding: 0px;
    margin:0!important;
}
.google-visualization-tooltip-item-list {
  	list-style-type: none;
    margin: 0!important;
    padding: 0!important;
}
.google-visualization-tooltip-item:first-child {
    margin: 0!important;
}
.google-visualization-tooltip-item {
    margin: 0!important;
    padding: 0!important;
}
