a.macro {
	font-family: SFMono-Regular,Menlo,Monaco,Consolas,"Liberation Mono","Courier New",Courier,monospace;
	font-size: 90%;
}

table th p {
	margin-bottom: 0px;
}

table.indextable {
	margin-bottom: 20px;
}

table.indextable li {
	max-width: 350px;
	margin-left: 10px;
	word-wrap: break-word;
}

.rst-content dl:not(.docutils) dt {
	display: table-cell;
}

nav.wy-nav-top {
	/* Transparent Condor Red with Black to match side bar theme color */
	background-color: #7a2b3a;
}

div.wy-side-nav-search {
	background-color: #d312456e;
}

div.wy-side-nav-search a.icon.icon-home:before {
	content: url(../logo.svg);
	width: 25px;
	height: 15px;
}

span.faux-definition {
	display: none;
}

/* Enable tool mindmap links */
[class*="link:"] {
	cursor: pointer;
}

/* Override Sidebar Section Header Colors*/
.wy-menu-vertical header, .wy-menu-vertical p.caption {
	color: #e69eb1;
}

/* override table width restrictions */
@media screen and (min-width: 767px) {
	.wy-table-responsive table td {
	   /* !important prevents the common CSS stylesheets from overriding
		  this as on RTD they are loaded after this stylesheet */
	   white-space: normal !important;
	}

	.wy-table-responsive {
	   overflow: visible !important;
	}

	.wy-nav-content {
	    max-width: 95%;
	}
 }

/* override the formatting of Python documentation links */
.rst-content code.xref.py.py-obj.docutils.literal.notranslate,
.rst-content code.xref.py.py-data.docutils.literal.notranslate,
.rst-content code.xref.py.py-meth.docutils.literal.notranslate,
.rst-content code.xref.py.py-class.docutils.literal.notranslate
{
    color: #000000;
    text-decoration: underline;
    border: none;
    background-color: transparent;
    font-weight: normal;
    font-size: 100%;
}

/* override the ``literal`` formating */
.rst-content code:not(.xref):not(.py):not(.py-data).docutils.literal.notranslate,
.rst-content   tt:not(.xref):not(.py):not(.py-data).docutils.literal.notranslate
{
    color: #000000;
    background-color: transparent;
    border: none;
    font-size: 90%;
    background-color: ivory;
}
